{
   uFile, file operations and abstraction
   Copyright (C) 2007. Dejan Boras

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   Started On:    14.12.2007.
   Last Update:   30.04.2011.
}

{$MODE OBJFPC}{$H+}{$I-}
UNIT uFile;

INTERFACE

   USES uStd, uError, StringUtils;

CONST
   {ERROR CONSTANTS}
   feOPENED          = $100; {cannot perform operation if file opened}
   feCLOSED          = $101; {cannot perform operation if file closed}
   feERROR_STATE     = $102; {file is in error state}
   feOPEN            = $103; {failed to open file}
   feNEW             = $104; {failed to create file}

   {FILE OPERATION CONSTANTS}
   {seek operations}
   fcSEEKSET      = $00; {seek from the position 0}
   fcSEEKCUR      = $01; {seek from the current position}
   fcSEEKEND      = $02; {seek from end}

   {FILE MODES}
   fcfNONE        = $0000; {no file mode specified}
   fcfREAD        = $0001; {read-only}
   fcfWRITE       = $0002; {write-only}
   fcfRW          = fcfREAD or fcfWRITE; {read and write}

   {LOGGING}
   fcLog: boolean = true;
   fcLogExtra: boolean = true;

TYPE
   {$IFDEF DFILEFS_INT32}
      fileint = longint;
   {$ELSE}
      fileint = int64;
   {$ENDIF}

   {routine types used by the file handler}
   fTMakeProc     = procedure(var f);
   fTDisposeProc  = procedure(var f);
   fTOpenProc     = procedure(var f);
   fTNewProc      = procedure(var f);
   fTReadFunc     = function (var f; var buf; count: fileint): fileint;
   fTWriteFunc    = function (var f; const buf; count: fileint): fileint;
   fTSeekProc     = function(var f; pos: fileint): fileint;
   fTCloseProc    = procedure(var f);
   fTOnBufferSetProc = procedure (var f);
   fTFlushProc    = procedure(var f);

   {a custom file handler}
   PFileHandler = ^TFileHandler;
   TFileHandler = record
      make:    fTMakeProc;
      dispose: fTDisposeProc;
      open:    fTOpenProc;
      new:     fTNewProc;
      read:    fTReadFunc;
      write:   fTWriteFunc;
      seek:    fTSeekProc;
      close:   fTCloseProc;
      flush:   fTFlushProc;
      onbufferset: fTOnBufferSetProc;

      useBuffering, doReadUp: boolean;
   end;

   {a OrcinusXDL file}
   PFile = ^TFile;
   TFile = record
      fn: string;
      fMode, fType, fNew: longword;

      fSize, fSizeLimit, fPosition, fOffset, {size of the file and position}
      bSize, bLimit, bPosition: fileint; {size of the buffer and position}

      pData: pointer;
      bData: pbyte;
      pHandler: PFileHandler;

      {extra data}
      extData: pointer;
      handle: longint;
      pSub: PFile;
   end;

{$IFNDEF FILE_NOFS}
TYPE
   PVFileSystem = ^TVFileSystem;
   TVFileSystem = record
      FileInFS: function(const path: string): fileint;
      OpenFile: function(var f: TFile; const fn: string): boolean;
      NewFile: function(var f: TFile; const fn: string): boolean;

      next: PVFileSystem;
   end;
{$ENDIF}

CONST
   {BUFFER SETTINGS}
   fcMinimumBufferSize = 1 * 1024; {minimal size of the buffer}
   fcCopyBufferSize: longint = 32 * 1024;
   fcAutoSetBuffer: boolean = true; {automatically set buffer for each file}
   fcAutoSetBufferSize: fileint = fcMinimumBufferSize; {auto set buffer size}

VAR
   {error variables}
   fError, fErrorD: longint;
   fDummyHandler: TFileHandler;

   {$IFNDEF FILE_NOFS}
   dummyFS: TVFileSystem;
   {$ENDIF}

{ERROR SUPPORT}
procedure feRaise(err, errd: longint);
procedure feRaise(err: longint);
procedure fErrorReset();
procedure fErrorIgn();

{GENERAL FILE OPERATIONS}

{set defaults for a file}
procedure fSetDefaults(var f: TFile; const fn: string);
{checks whether a file exists, returns filesize if the file exists, or -1 on error}
function fExist(const fn: string): fileint;
{return the size of the file}
function fSize(var f: TFile): fileint;
{erase a file with the filename fn}
procedure fErase(const fn: string);

{FILE OPERATIONS}

{GENERAL}
procedure fInit(var f: TFile); inline;
procedure fInit(var h: TFileHandler); inline;
procedure fDispose(var f: TFile);

{FILE HANDLING}

{set the mode in which the file will be opened}
procedure fFileMode(fm: longword);

{FILE HANDLER}
{assign a file handler to a file}
procedure fAssignHandler(var f: TFile; var handler: TFileHandler);

{read an entire file into memory once it is opened, only works for reading mode}
procedure fReadUp(var f: TFile);
{close a opened file}
procedure fClose(var f: TFile);
{set a file buffer}
procedure fBuffer(var f: TFile; Size: fileint);
{dispose a file buffer}
procedure fDisposeBuffer(var f: TFile);

{set the file mode}
procedure fMode(var f: TFile; mode: longword);

{READING AND WRITING}
{read count bytes into buffer buf from file f}
function fRead(var f: TFile; var buf; count: fileint): fileint;
{write count bytes from buffer buf to file f}
function fWrite(var f: TFile; const buf; count: fileint): fileint;
{flushes contents of buffer to the file if there is anything to write}
procedure fFlush(var f: TFile);
{a kludge used to read a line from a file, even a binary one}
procedure fReadln(var f: TFile; var s: string);
{seeks to a position in the file}
function fSeek(var f: TFile; position: fileint): fileint;
function fSeek(var f: TFile; position: fileint; mode: longword): fileint;
{checks if the file has reached the end}
function fEOF(var f: TFile): boolean;

{ OTHER }
{save the contents of memory to a file}
function fSaveMem(const fn: string; var m; size: fileint): fileint;
{load the contents of a file into memory}
function fLoadToMem(const fn: string; var m: pointer): fileint;

{ ADDITIONAL }
function fReadAnsiString(var f: TFile; var s: ansistring): fileint;
function fWriteAnsiString(var f: TFile; const s: ansistring): fileint;

{ FILESYSTEM }
{$IFNDEF FILE_NOFS}
procedure fsInit(var fs: TVFileSystem);
procedure fsAdd(var fs: TVFileSystem);
function fsExists(const fn: string): fileint;
function fsOpen(var f: TFile; const fn: string): boolean;
{$ENDIF}

IMPLEMENTATION

{$IFNDEF NO_LOGGING}
{LOGGING HANDLING}
function LogOk(): boolean;
begin
   LogOk := false;
   if(logOk = true) and (fcLog = true) then LogOk := true;
end;

function LogExtraOk(): boolean;
begin
   LogExtraOk := false;
   if(logOk = true) and (fcLog = true)
      and (fcLogExtra = true)then LogExtraOk := true;
end;
{$ENDIF}

{ERROR HANDLING}
procedure feRaise(err, errd: longint);
begin
   fError := err; fErrorD := errd;
end;

procedure feRaise(err: longint);
begin
   feRaise(err, 0);
end;

procedure fErrorReset();
begin
   fError := 0; fErrorD := 0; ioE := 0;
end;

procedure fErrorIgn();
begin
   fError := 0; ioErrorIgn();
end;

{GENERAL FILE OPERATIONS}

procedure fSetDefaults(var f: TFile; const fn: string);
begin
   f.fPosition := 0;
   f.bPosition := 0;
   f.bLimit := 0;
   f.fn := fn;
end;

function fExist(const fn: string): fileint;
var
   f: file;
   found: boolean = true;

begin
   result := -1;

   {first try regular files}
   Assign(f, fn);
   Reset(f, 1);
   if(ioerror <> 0) then found := false;

   result := FileSize(f);
   ioerror();

   Close(f);
   ioerror();

   {$IFNDEF FILE_NOFS}
   if(not found) then
      result := fsExists(fn);
   {$ENDIF}
end;

function fSize(var f: TFile): fileint;
begin
   result := f.fSize;
end;

procedure fErase(const fn: string);
var
   f: file;

begin
   Assign(f, fn);
   Erase(f);
   if(ioerror <> 0) then feRaise(eIO);
end;

{FILE OPERATIONS}

{GENERAL}
procedure fInit(var f: TFile); inline;
begin
   Zero(f, SizeOf(TFile));
end;

procedure fInit(var h: TFileHandler); inline;
begin
   h := fDummyHandler;
end;

procedure fHandlerDispose(var f: TFile);
begin
   if(f.pHandler <> nil) then
      if(f.pHandler^.Dispose <> nil) then f.pHandler^.Dispose(f);
end;

procedure fDispose(var f: TFile);
begin
   fDisposeBuffer(f);
   fHandlerDispose(f);
end;

{FILE HANDLING}
VAR
   fileMode: longword;

{set the mode in which the file will be opened}
procedure fFileMode(fm: longword);
begin
   fileMode := fm;
   if(fileMode > $02) then feRaise(eInvalid);
end;

{assign a file handler to a file}
procedure fAssignHandler(var f: TFile; var handler: TFileHandler);
begin
   fHandlerDispose(f);

   f.pHandler := @handler;
   handler.make(f);
end;

{read an entire file into memory once it is opened, only works for reading mode}
procedure fReadUp(var f: TFile);
var
   doReadUp: boolean;

begin
   fErrorReset();
   if(f.pHandler <> nil) then
      doReadUp := f.pHandler^.doReadUp
   else doReadUp := true;


   if(doReadUp = true) then begin
      fBuffer(f, f.fSize);
      if(fError = 0) then begin
         fSeek(f, 0);
         if(fError = 0) then
            fRead(f, f.bData^, f.fSize);
      end;
   end;
end;

{close a opened file}
procedure fClose(var f: TFile);
begin
   fFlush(f);

   fSeek(f, 0);

   if(f.pHandler <> nil) then
      if(f.pHandler^.close <> nil) then f.pHandler^.close(f);

   fHandlerDispose(f);
end;

{dispose a file buffer}
procedure fDisposeBuffer(var f: TFile);
begin
   fFlush(f);

   XFreeMem(f.bData); f.bSize := 0;
   if(f.pHandler <> nil) then
      if(f.pHandler^.onbufferset <> nil) then f.pHandler^.onbufferset(f);
end;

{set a file buffer}
procedure fBuffer(var f: TFile; Size: fileint);
begin
   if(Size > 0) and (f.fMode <> fcfRW) then begin
      if(Size < fcMinimumBufferSize) then Size := fcMinimumBufferSize;

      fDisposeBuffer(f);
      GetMem(f.bData, Size);
      if(f.bData <> nil) then begin
         f.bSize := Size;
         f.bPosition := 0;
         f.bLimit := 0;
      end else feRaise(eNO_MEMORY);

      {call handler}f.pHandler^.onbufferset(f);
   end else fDisposeBuffer(f);
end;

{set the file mode}
procedure fMode(var f: TFile; mode: longword);
begin
   f.fMode := mode;
end;

{READING AND WRITING}
{read count bytes into buffer buf from file f}
function fRead(var f: TFile; var buf; count: fileint): fileint;
var
   bRead: fileint; {how much was read, how much is left in buffer}

begin
   if(f.fMode and fcfREAD > 0) then begin
      if(f.fSizeLimit <> 0) then
         if(f.fSizeLimit < f.fPosition+count) then count := f.fSize - f.fPosition
      else
         if(f.fSize < f.fPosition+count) then count := f.fSize - f.fPosition;

      bRead := f.pHandler^.Read(f, buf, count);
      if(fError = 0) then
         inc(f.fPosition, bRead);

      result := bRead;
   end else begin
      feRaise(eREAD); result := -1;
   end;
end;

function fWrite(var f: TFile; const buf; count: fileint): fileint;
var
   bWrite: fileint;

begin
   if(f.fMode and fcfWRITE > 0) then begin
      if(f.fSizeLimit <> 0) then
         if(f.fSizeLimit < f.fPosition+count) then count := f.fSize - f.fPosition;

      bWrite := f.pHandler^.Write(f, buf, count);
      if(fError = 0) then
         inc(f.fPosition, bWrite);
      result := bWrite;
   end else begin
      result := -1; feRaise(eWRITE);
   end;
end;

procedure fFlush(var f: TFile);
begin
   {does the file have a buffer}
   if(f.bSize > 0) and (f.bPosition > 0) and (f.fMode = fcfWRITE) then begin
      if(f.pHandler <> nil) then
         if(f.pHandler^.flush <> nil) then
            f.pHandler^.flush(f);
   end;
end;

procedure fReadln(var f: TFile; var s: string);
var
   c: char = #0;
   chars: longint = 0;
   count: fileint = 0;

begin
   SetLength(s, 0);
   chars := 0;

   repeat
      count := fRead(f, c, 1);
      if(fError = 0) and (count > 0) then begin
         if(c = #10) and (chars = 0) then continue;
         if (c <> #13) then begin
            inc(chars);
            SetLength(s, chars);
            s[chars] := c;
         end;
      end else break;
   until (c = #13);
end;

function fSeek(var f: TFile; position: fileint): fileint;
begin
   fSeek := -1;

   if(position <= f.fSize) then begin
      result := f.pHandler^.seek(f, position);
      if(fError = 0) then begin
         f.fPosition := position;
         f.bPosition := 0; {invalidate buffer}
         f.bLimit := 0;
      end;
   end;
end;

function fSeek(var f: TFile; position: fileint; mode: longword): fileint;
begin
   case mode of
      fcSEEKSET: result := fSeek(f, Position);
      fcSEEKCUR: result := fSeek(f, f.fPosition+Position);
      fcSEEKEND: result := fSeek(f, f.fSize-Position);
      else fSeek := -1;
   end;
end;

function fEOF(var f: TFile): boolean;
begin
   result := false;
   if((f.fPosition-f.bPosition) < f.fSize) then result := false
   else result := true;
end;

{DUMMY FILE HANDLER}
{$PUSH}{$HINTS OFF}{these 'dumb' routines don't need to use any parameters}
procedure dmbProc    (var f); begin end;
function  dmbRead    (var f; var buf; count: fileint): fileint; begin dmbRead := 0; end;
function  dmbWrite   (var f; var buf; count: fileint): fileint; begin dmbWrite := 0; end;
function  dmbSeek    (var f; pos: fileint): fileint; begin dmbSeek := 0; end;
{$POP}

procedure InitFileHandlers();
begin
   {dummy file handler}
   fDummyHandler.make          := @dmbProc;
   fDummyHandler.dispose       := @dmbProc;
   fDummyHandler.open          := @dmbProc;
   fDummyHandler.new           := @dmbProc;
   fDummyHandler.read          := ftReadFunc     (@dmbRead);
   fDummyHandler.write         := ftWriteFunc    (@dmbWrite);
   fDummyHandler.Seek          := ftSeekProc     (@dmbSeek);
   fDummyHandler.close         := @dmbProc;
   fDummyHandler.onbufferset   := @dmbProc;
   fDummyHandler.useBuffering  := true;
   fDummyHandler.doReadUp      := true;
end;

{ OTHER }
function fSaveMem(const fn: string; var m; size: fileint): fileint;
var
   f: file;
   brw: fileint = 0;

begin
   fSaveMem := -1;

   {open the file}
   Assign(f, fn);
   Rewrite(f, 1);
   if(ioerror <> 0) then begin feRaise(eIO); exit; end;

   {save the memory contents to the file}
   blockwrite(f, m, size, brw);
   if(ioerror <> 0) then begin
      feRaise(eIO); close(f); ioErrorIgn(); exit;
   end;

   {close the file}
   close(f);
   if(ioerror <> 0) then begin feRaise(eIO); exit; end;

   fSaveMem := brw;
end;

function fLoadToMem(const fn: string; var m: pointer): fileint;
var
   f: file;
   br: fileint = 0;
   size: fileint;

procedure cleanup();
begin
   close(f); ioErrorIgn();
end;

begin
   fLoadToMem := -1;

   {open the file}
   Assign(f, fn);
   Reset(f, 1);
   if(ioerror <> 0) then begin feRaise(eIO); exit; end;

   {get the file size}
   size := FileSize(f);
   if(ioerror <> 0) then begin
      feRaise(eIO); cleanup(); exit;
   end;

   {allocate memory for the file}
   GetMem(m, size);
   if(m <> nil) then begin
      {read the file contents to the memory}
      blockread(f, m^, size, br);
      if(ioerror = 0) then begin
         {close the file}
         close(f);
         if(ioerror = 0) then
            {return how much was read}
            fLoadToMem := br
         else feRaise(eIO);
      end else feRaise(eIO);
   end else feRaise(eNO_MEMORY);
   cleanup();
end;


{ ADDITIONAL }

function fReadAnsiString(var f: TFile; var s: ansistring): fileint;
var
   len, bread: longint;

begin
   {read string length}
   bRead := fRead(f, len, SizeOf(len));
   if(fError = 0) then begin
      SetLength(s, len);
      if(len > 0) then begin
         {read string contents}
         inc(bRead, fRead(f, s[1], len));
      end;
   end;

   result := bRead;
end;

function fWriteAnsiString(var f: TFile; const s: ansistring): fileint;
var
   len, bWrote: longint;

begin
   len := Length(s);
   {write string length}
   bWrote := fWrite(f, len, SizeOf(len));
   {write string contents}
   if(fError = 0) and (len > 0) then
      inc(bWrote, fWrite(f, s[1], len));

   result := bWrote;
end;

{ FILESYSTEM }
{$IFNDEF FILE_NOFS}
VAR
   fss, fse: PVFileSystem;

procedure fsInit(var fs: TVFileSystem);
begin
   fs := dummyFS;
end;

procedure fsAdd(var fs: TVFileSystem);
begin
   fs.next := nil;
   if(fss = nil) then fss := @fs
   else fse^.next := @fs;
   fse := @fs;
end;

function fsExists(const fn: string): fileint;
var
   cur: PVFileSystem;
   res: fileint;

begin
   result := -1;

   cur := fss;
   if(cur <> nil) then repeat
      res := cur^.FileInFS(fn);
      if(res > -1) then exit(res);

      cur := cur^.next;
   until (cur = nil);
end;

function fsOpen(var f: TFile; const fn: string): boolean;
var
   cur: PVFileSystem;

begin
   result := false;

   cur := fss;
   if(cur <> nil) then repeat
      if(cur^.FileInFS(fn) > -1) then
         exit(cur^.OpenFile(f, fn));

      cur := cur^.next;
   until (cur = nil);
end;

{ DUMMY FS }
{$PUSH}{$HINTS OFF} {since these are dummy routines most parameters are unused}
function dumFileInFS(const path: string): fileint; begin result := -1; end;
function dumOpenFile(var f: TFile; const fn: string): boolean; begin result := false; end;
function dumNewFile(var f: TFile; const fn: string): boolean; begin result := false; end;
{$POP}
{$ENDIF}

INITIALIZATION
   {$IFNDEF FILE_NOFS}
   dummyFS.FileInFS := @dumFileInFS;
   dummyFS.OpenFile := @dumOpenFile;
   dummyFS.NewFile := @dumNewFile;
   {$ENDIF}

   InitFileHandlers();
END.
